# Author: Zylo117

import math

import torch
from torch import nn

from .efficientdet.model import BiFPN, Center_layer, Proto_layer, Coef_layer,Offset_layer, Heatmap_layer, EfficientNet, SeparableConvBlock


class EfficientDetBackbone(nn.Module):
    def __init__(self, num_classes=1, compound_coef=0, load_weights=False, **kwargs):
        super(EfficientDetBackbone, self).__init__()
        self.compound_coef = compound_coef

        self.backbone_compound_coef = [0, 1, 2, 3, 4, 5, 6, 6]
        self.fpn_num_filters = [64, 88, 112, 160, 224, 288, 384, 384]
        self.fpn_cell_repeats = [3, 4, 5, 6, 7, 7, 8, 8]
        self.input_sizes = [512, 640, 768, 896, 1024, 1280, 1280, 1536]
        self.box_class_repeats = [3, 3, 3, 4, 4, 4, 5, 5]
        self.anchor_scale = [4., 4., 4., 4., 4., 4., 4., 5.]
        self.aspect_ratios = kwargs.get('ratios', [(1.0, 1.0), (1.4, 0.7), (0.7, 1.4)])
        self.num_scales = len(kwargs.get('scales', [2 ** 0, 2 ** (1.0 / 3.0), 2 ** (2.0 / 3.0)]))
        conv_channel_coef = {
            # the channels of P3/P4/P5.
            0: [40, 112, 320],
            1: [40, 112, 320],
            2: [48, 120, 352],
            3: [48, 136, 384],
            4: [56, 160, 448],
            5: [64, 176, 512],
            6: [72, 200, 576],
            7: [72, 200, 576],
        }

        num_anchors = len(self.aspect_ratios) * self.num_scales

        self.bifpn = nn.Sequential(
            *[BiFPN(self.fpn_num_filters[self.compound_coef],
                    conv_channel_coef[compound_coef],
                    True if _ == 0 else False,
                    attention=True if compound_coef < 6 else False)
              for _ in range(self.fpn_cell_repeats[compound_coef])])

        self.conv = SeparableConvBlock(self.fpn_num_filters[self.compound_coef])
        self.upsample = nn.Upsample(scale_factor=2, mode='nearest')
        self.coord_channel = self.CoordChannel
        self.num_classes = num_classes
        # self.proto_head = Proto_layer(in_channels=self.fpn_num_filters[self.compound_coef])
        # self.coef_head = Coef_layer(in_channels=self.fpn_num_filters[self.compound_coef])
        self.center = Center_layer(in_channels=self.fpn_num_filters[self.compound_coef],num_classes=num_classes)
        self.offset_layer=Offset_layer(in_channels=self.fpn_num_filters[self.compound_coef])

        self.backbone_net = EfficientNet(self.backbone_compound_coef[compound_coef], load_weights)


    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()


    def CoordChannel(self, feature):

        # 将x,y坐标拼接在feature map上,通道数+2,
        # 这里是将坐标信息concat到feature map上
        x_range = torch.linspace(-1, 1, feature.shape[-1], device=feature.device) # w --> x
        y_range = torch.linspace(-1, 1, feature.shape[-2], device=feature.device) # h --> y
        # 对 x_range, y_range 进行扩充 
        y, x = torch.meshgrid(y_range, x_range)
        # 将两个坐标扩成4维
        y = y.expand([feature.shape[0], 1, -1, -1])
        x = x.expand([feature.shape[0], 1, -1, -1])
        # 将坐标cancat到feature map的通道上
        coord_feature = torch.cat([x, y], 1)
        ins_feature = torch.cat([feature, coord_feature], 1)
        
        return ins_feature


    def forward(self, inputs):
        max_size = inputs.shape[-1]

        _, p3, p4, p5 = self.backbone_net(inputs)

        features = [p3, p4, p5]
        features = self.bifpn(features)

        feature = self.conv(self.upsample(features[0]))
        centermap = self.center(feature)
        offsets=self.offset_layer(feature)
        return centermap,offsets


    def init_backbone(self, path):
        state_dict = torch.load(path)
        try:
            ret = self.load_state_dict(state_dict, strict=False)
            print(ret)
        except RuntimeError as e:
            print('Ignoring ' + str(e) + '"')


if __name__ == '__main__':
    input_tensor = torch.ones((4, 3, 512, 512))
    model = EfficientDetBackbone()
    centermap = model(input_tensor)
    print(centermap.shape)



from .network import CircleNet,CircleDetCriterion
from .dataset import CircleDetDataset
from .predict import CircleDetector
from .training import train
from .testing import test
from .utils import visualize
from .config import CircleDetConfig



import xml.etree.ElementTree as ET
def read_pascalvoc_annotation(file,cls2idx):
    root  = ET.parse(open(file,'r')).getroot()

    size = root.find('size')
    # w = int(size.find('width').text)
    # h = int(size.find('height').text)
    # path = root.find('path').text
    objects = []
    for obj in root.iter('object'):
        cls = obj.find('name').text
        if cls not in cls2idx:
            continue
        cls_id = cls2idx[cls]
        xmlbox = obj.find('bndbox')
        b = (float(xmlbox.find('xmin').text), float(xmlbox.find('ymin').text), float(xmlbox.find('xmax').text),
             float(xmlbox.find('ymax').text))
        obj = [list(b), cls_id]
        objects.append(obj)
    return objects


import torch
import os
import copy

# class ConfigOld:
#     class __empty_:pass
#     @classmethod
#     def _to_dict(cls,recursive=True,soft_merge=False):
#         '''soft_merge: try to select and non-None value from [cls,cls.__base__,...]'''
#         if recursive and issubclass(cls.__base__,Config):
#             dic=cls.__base__._to_dict(recursive=recursive,soft_merge=soft_merge)
#         else:
#             dic={}
#         for k,v in cls.__dict__.items():
#             if not k.startswith('_'):
#                 if soft_merge and v is None:
#                     v_old=dic.get(k,None)
#                     if v_old is not None:
#                         v=v_old
#                 dic[k]=v
#         return dic
#     @classmethod
#     def _get(cls,key,default=__empty_):
#         if hasattr(cls,key):
#             return getattr(cls,key)
#         if issubclass(cls.__base__,Config):
#             return cls.__base__._get(key,default)
#         else:
#             if default is Config.__empty_:
#                 raise AttributeError('Config class %s has no attribute %s.'%(cls,key))
#             else:
#                 return default
# class config_metaclass(type):
#     def __new__(cls, name,bases,attrs):
#         base=bases[0] if len(bases) else None
#         if base and hasattr(base,'__attr_dict__'):
#             attr_dict=base.__attr_dict__.copy()
#         else:
#             attr_dict={}
#         for k,v in attrs.items():
#             if not k.startswith('__'):
#                 attr_dict[k]=v
#         attrs['__attr_dict__']=attr_dict
#         return type.__new__(cls,name,bases,attrs)
# class Config(metaclass=config_metaclass):
#     def __init__(self,**kwargs):
#         self.__attrs_dict__=copy.deepcopy(self.__class__.__attrs_dict__)
#         self.__attrs_dict__.update(**kwargs)
#     @classmethod
#     def __to_dict__(cls):
#         return cls.__attr_dict__.copy()
#
# class DemoBase(object):
#     class __config__(Config):
#         BATCH_SIZE=4
#         INPUT_SIZE=(512,512)
#     class __allowed_options__(Config):
#         BATCH_SIZE=None
#
#     def __init__(self,cfg=None,**kwargs):
#         self.set_attrs(self.__config__.__attrs_dict__.cppy())
#         if cfg:
#             pass
#         self.set_attrs(kwargs)
#     def set_attrs(self,attrs):
#         for k,v in attrs.items():
#             setattr(self,k,v)
#
#
# class Demo(DemoBase,object):
#     class __config__(DemoBase.__config__):
#         BATCH_SIZE = 4


# print(Demo.__config__.__to_dict__())

# print(Demo.__bases__)




class ConfigBase:
    __required__ = []
    __methods__=['init','update','train','test']
    def __init__(self, init=True, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.check_required()
        if init:
            self.init()

    def init(self):
        pass

    def update(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def check_required(cls):
        for key in cls.__required__:
            if getattr(cls, key, None) is None:
                raise Exception('Argument %s is required but not given.' % (key))

    @classmethod
    def to_dict(cls):
        if issubclass(cls.__base__,ConfigBase):
            dic=cls.__base__.to_dict()
        else:
            dic = {}
        for k, v in cls.__dict__.items():
            if not (k.startswith('__') and k.endswith('__') and k not in cls.__methods__):
                dic[k] = v
        return dic


class ConfigHintBase(ConfigBase):
    INPUT_SIZE = None
    MAX_EPOCHS = None
    CLASSES = None
    NUM_CLASSES = None
    NUM_POINTS = None
    NUM_REG_PARAMS=None
    BATCH_SIZE = None
    LEARN_RATE_INIT = None
    LEARN_RATE_END = None
    WEIGHTS_SAVE_INTERVAL = None
    WEIGHTS_DIR = None
    WEIGHTS_INIT = None
    DEVICE = None
    train_transform = None
    criterion = None
    get_model = None
    get_train_data = None
    get_val_data = None
    get_predictor = None
    TRAIN_DIR = None
    VAL_DIR = None
    TEST_DIR = None
    DATA_DIR = None
    DET_TYPE=None
    __required__ = []


class DetConfigBase(ConfigHintBase):
    BATCH_SIZE = 4
    INPUT_SIZE = (512, 512)
    MAX_EPOCHS = 200
    LEARN_RATE_INIT = 1e-4
    LEARN_RATE_END = 1e-6
    WEIGHTS_SAVE_INTERVAL = 20
    WEIGHTS_DIR = 'weights/training'
    WEIGHTS_INIT = None
    criterion = None
    DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    __required__ = ['get_model', 'criterion']
    __dataset__ = None
    __model__ = None

    def get_dataset(self):
        raise NotImplementedError

    def get_model(self):
        raise NotImplementedError

    def get_detector(self):
        raise NotImplementedError

    def test(self):
        raise NotImplementedError

    def train(self):
        raise NotImplementedError

    def init(self):
        if not os.path.exists(self.WEIGHTS_DIR):
            os.makedirs(self.WEIGHTS_DIR)




import numpy as np
import cv2
import random


def flip(img):
	return img[:, :, ::-1].copy()


def transform_preds(coords, center, scale, output_size):
	target_coords = np.zeros(coords.shape)
	trans = get_affine_transform(center, scale, 0, output_size, inv=1)
	for p in range(coords.shape[0]):
		target_coords[p, 0:2] = affine_transform(coords[p, 0:2], trans)
	return target_coords


def get_affine_transform(center,
                         scale,
                         rot,
                         output_size,
                         shift=np.array([0, 0], dtype=np.float32),
                         inv=0):
	if not isinstance(scale, np.ndarray) and not isinstance(scale, list):
		scale = np.array([scale, scale], dtype=np.float32)

	scale_tmp = scale
	src_w = scale_tmp[0]
	dst_w = output_size[0]
	dst_h = output_size[1]

	rot_rad = np.pi * rot / 180
	src_dir = get_dir([0, src_w * -0.5], rot_rad)
	dst_dir = np.array([0, dst_w * -0.5], np.float32)

	src = np.zeros((3, 2), dtype=np.float32)
	dst = np.zeros((3, 2), dtype=np.float32)
	src[0, :] = center + scale_tmp * shift
	src[1, :] = center + src_dir + scale_tmp * shift
	dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
	dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5], np.float32) + dst_dir

	src[2:, :] = get_3rd_point(src[0, :], src[1, :])
	dst[2:, :] = get_3rd_point(dst[0, :], dst[1, :])

	if inv:
		trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
	else:
		trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))

	return trans


def affine_transform(pt, t):
	new_pt = np.array([pt[0], pt[1], 1.], dtype=np.float32).T
	new_pt = np.dot(t, new_pt)
	return new_pt[:2]


def get_3rd_point(a, b):
	direct = a - b
	return b + np.array([-direct[1], direct[0]], dtype=np.float32)


def get_dir(src_point, rot_rad):
	sn, cs = np.sin(rot_rad), np.cos(rot_rad)

	src_result = [0, 0]
	src_result[0] = src_point[0] * cs - src_point[1] * sn
	src_result[1] = src_point[0] * sn + src_point[1] * cs

	return src_result


def crop(img, center, scale, output_size, rot=0):
	trans = get_affine_transform(center, scale, rot, output_size)

	dst_img = cv2.warpAffine(img,
	                         trans,
	                         (int(output_size[0]), int(output_size[1])),
	                         flags=cv2.INTER_LINEAR)

	return dst_img


def gaussian_radius(det_size, min_overlap=0.7):
	height, width = det_size

	a1 = 1
	b1 = (height + width)
	c1 = width * height * (1 - min_overlap) / (1 + min_overlap)
	sq1 = np.sqrt(b1 ** 2 - 4 * a1 * c1)
	r1 = (b1 + sq1) / 2

	a2 = 4
	b2 = 2 * (height + width)
	c2 = (1 - min_overlap) * width * height
	sq2 = np.sqrt(b2 ** 2 - 4 * a2 * c2)
	r2 = (b2 + sq2) / 2

	a3 = 4 * min_overlap
	b3 = -2 * min_overlap * (height + width)
	c3 = (min_overlap - 1) * width * height
	sq3 = np.sqrt(b3 ** 2 - 4 * a3 * c3)
	r3 = (b3 + sq3) / 2
	return min(r1, r2, r3)


def gaussian2D(shape, sigma=1):
	m, n = [(ss - 1.) / 2. for ss in shape]
	y, x = np.ogrid[-m:m + 1, -n:n + 1]

	h = np.exp(-(x * x + y * y) / (2 * sigma * sigma))
	h[h < np.finfo(h.dtype).eps * h.max()] = 0
	return h



def draw_points_heatmap(heatmap,points, radius=10):

	for point in points:
		radius=gaussian_radius_cornernet((25,25),0.7)
		heatmap = draw_umich_gaussian_1(heatmap, point, int(radius),int(radius))
	return heatmap
def gaussian2D_1(shape, sigmah=1, sigmaw=1):
	m, n = [(ss - 1.) / 2. for ss in shape]
	y, x = np.ogrid[-m:m + 1, -n:n + 1]
	h = np.exp(-((x * x) / (2 * sigmah * sigmaw) + (y * y) / (2 * sigmah * sigmah)))
	h[h < np.finfo(h.dtype).eps * h.max()] = 0
	return h
def draw_umich_gaussian_1(heatmap, center, rh, rw, k=1):
	diameterh = 2 * rh + 1
	diameterw = 2 * rw + 1
	gaussian = gaussian2D_1((diameterh, diameterw), sigmah=diameterh / 6, sigmaw=diameterw / 6)

	x, y = int(center[0]), int(center[1])
	height, width = heatmap.shape[0:2]

	left, right = min(x, rw), min(width - x, rw + 1)
	top, bottom = min(y, rh), min(height - y, rh + 1)

	masked_heatmap = heatmap[y - top:y + bottom, x - left:x + right]
	masked_gaussian = gaussian[rh - top:rh + bottom, rw - left:rw + right]
	if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:
		np.maximum(masked_heatmap, masked_gaussian * k, out=masked_heatmap)
	return heatmap
def gaussian_radius_cornernet(det_size, min_overlap):
	height, width = det_size

	a1 = 1
	b1 = (height + width)
	c1 = width * height * (1 - min_overlap) / (1 + min_overlap)
	sq1 = np.sqrt(b1 ** 2 - 4 * a1 * c1)
	r1 = (b1 - sq1) / (2 * a1)

	a2 = 4
	b2 = 2 * (height + width)
	c2 = (1 - min_overlap) * width * height
	sq2 = np.sqrt(b2 ** 2 - 4 * a2 * c2)
	r2 = (b2 - sq2) / (2 * a2)

	a3 = 4 * min_overlap
	b3 = -2 * min_overlap * (height + width)
	c3 = (min_overlap - 1) * width * height
	sq3 = np.sqrt(b3 ** 2 - 4 * a3 * c3)
	r3 = (b3 + sq3) / (2 * a3)
	return min(r1, r2, r3)


def draw_umich_gaussian(heatmap, center, radius, k=1):
	radius = int(radius)
	diameter = 2 * radius + 1
	gaussian = gaussian2D((diameter, diameter), sigma=diameter / 6)

	x, y = int(center[0]), int(center[1])

	height, width = heatmap.shape[0:2]

	left, right = min(x, radius), min(width - x, radius + 1)
	top, bottom = min(y, radius), min(height - y, radius + 1)
	left, right, top, bottom = [int(x) for x in [left, right, top, bottom]]

	masked_heatmap = heatmap[y - top:y + bottom, x - left:x + right]
	masked_gaussian = gaussian[radius - top:radius + bottom, radius - left:radius + right]
	if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:  # TODO debug
		np.maximum(masked_heatmap, masked_gaussian * k, out=masked_heatmap)
	return heatmap


def draw_dense_reg(regmap, heatmap, center, value, radius, is_offset=False):
	diameter = 2 * radius + 1
	gaussian = gaussian2D((diameter, diameter), sigma=diameter / 6)
	value = np.array(value, dtype=np.float32).reshape(-1, 1, 1)
	dim = value.shape[0]
	reg = np.ones((dim, diameter * 2 + 1, diameter * 2 + 1), dtype=np.float32) * value
	if is_offset and dim == 2:
		delta = np.arange(diameter * 2 + 1) - radius
		reg[0] = reg[0] - delta.reshape(1, -1)
		reg[1] = reg[1] - delta.reshape(-1, 1)

	x, y = int(center[0]), int(center[1])

	height, width = heatmap.shape[0:2]

	left, right = min(x, radius), min(width - x, radius + 1)
	top, bottom = min(y, radius), min(height - y, radius + 1)

	masked_heatmap = heatmap[y - top:y + bottom, x - left:x + right]
	masked_regmap = regmap[:, y - top:y + bottom, x - left:x + right]
	masked_gaussian = gaussian[radius - top:radius + bottom,
	                  radius - left:radius + right]
	masked_reg = reg[:, radius - top:radius + bottom,
	             radius - left:radius + right]
	if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:  # TODO debug
		idx = (masked_gaussian >= masked_heatmap).reshape(
			1, masked_gaussian.shape[0], masked_gaussian.shape[1])
		masked_regmap = (1 - idx) * masked_regmap + idx * masked_reg
	regmap[:, y - top:y + bottom, x - left:x + right] = masked_regmap
	return regmap


def draw_msra_gaussian(heatmap, center, sigma):
	tmp_size = sigma * 3
	mu_x = int(center[0] + 0.5)
	mu_y = int(center[1] + 0.5)
	w, h = heatmap.shape[0], heatmap.shape[1]
	ul = [int(mu_x - tmp_size), int(mu_y - tmp_size)]
	br = [int(mu_x + tmp_size + 1), int(mu_y + tmp_size + 1)]
	if ul[0] >= h or ul[1] >= w or br[0] < 0 or br[1] < 0:
		return heatmap
	size = 2 * tmp_size + 1
	x = np.arange(0, size, 1, np.float32)
	y = x[:, np.newaxis]
	x0 = y0 = size // 2
	g = np.exp(- ((x - x0) ** 2 + (y - y0) ** 2) / (2 * sigma ** 2))
	g_x = max(0, -ul[0]), min(br[0], h) - ul[0]
	g_y = max(0, -ul[1]), min(br[1], w) - ul[1]
	img_x = max(0, ul[0]), min(br[0], h)
	img_y = max(0, ul[1]), min(br[1], w)
	heatmap[img_y[0]:img_y[1], img_x[0]:img_x[1]] = np.maximum(
		heatmap[img_y[0]:img_y[1], img_x[0]:img_x[1]],
		g[g_y[0]:g_y[1], g_x[0]:g_x[1]])
	return heatmap


def grayscale(image):
	return cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)


def lighting_(data_rng, image, alphastd, eigval, eigvec):
	alpha = data_rng.normal(scale=alphastd, size=(3,))
	image += np.dot(eigvec, eigval * alpha)


def blend_(alpha, image1, image2):
	image1 *= alpha
	image2 *= (1 - alpha)
	image1 += image2


def saturation_(data_rng, image, gs, gs_mean, var):
	alpha = 1. + data_rng.uniform(low=-var, high=var)
	blend_(alpha, image, gs[:, :, None])


def brightness_(data_rng, image, gs, gs_mean, var):
	alpha = 1. + data_rng.uniform(low=-var, high=var)
	image *= alpha


def contrast_(data_rng, image, gs, gs_mean, var):
	alpha = 1. + data_rng.uniform(low=-var, high=var)
	blend_(alpha, image, gs_mean)


def color_aug(data_rng, image, eig_val, eig_vec):
	functions = [brightness_, contrast_, saturation_]
	random.shuffle(functions)

	gs = grayscale(image)
	gs_mean = gs.mean()
	for f in functions:
		f(data_rng, image, gs, gs_mean, 0.4)
	lighting_(data_rng, image, 0.1, eig_val, eig_vec)



import torch
import numpy as np
import os, glob, json, random
from torch.utils.data import Dataset
import cv2
from detro import utils as ut
from .det_dataset import DetDataset


class ShapeDetDataset(DetDataset):
    HEATMAP_DOWNSAMPLE = 4
    NUM_REG_PARAMS = None
    REG_MODE = 'offsets'

    __required__ = DetDataset.__required__ + ['CLASSES', 'HEATMAP_DOWNSAMPLE', 'transform','NUM_REG_PARAMS']

    def data_augmentation(self, img, objs):
        objs.sort(key=lambda obj: np.array(obj[0])[:, 0].mean())
        shapes = [ut.organize_polygon_points(shape) for shape, cls in objs]
        clses = [cls for shape, cls in objs]
        if self.transform:
            img, shapes = self.transform(img, shapes)
        objs = list(zip(shapes, clses))
        return img, objs

    def transform_points(self, points):
        mode = self.REG_MODE
        points = np.array(points)
        center = points.mean(axis=0)
        if mode == 'offsets-polar':
            points = ut.organize_polygon_points(points)
            return center, ut.cartesian2polar(points - center, to_degree=True).reshape((-1))
        if mode == 'fcos':
            points = ut.bounding_rect(points).reshape((2, 2))
            center = points.mean(axis=0)
            return center, np.abs(points - center).reshape(-1)
        if mode == 'centernet':
            p1, p2 = points = ut.bounding_rect(points).reshape((2, 2))
            w, h = abs(p1[0] - p2[0]), abs(p1[1] - p2[1])
            center = points.mean(axis=0)
            return center, np.array([w, h])
        if mode == 'circle':
            points = ut.organize_polygon_points(points)
            center = points.mean(axis=0)
            radius = 0.5 * ut.polygon_length(points) / len(points)
            return center, radius
        if mode == 'offsets':
            points = ut.organize_polygon_points(points)
            return center, (points - center).reshape((-1))
        else:
            raise

    def process_batch_annots(self, batch_annots):
        down = self.HEATMAP_DOWNSAMPLE
        batch_size = len(batch_annots)
        imw, imh = self.INPUT_SIZE
        hm_h, hm_w = int(imh / down), int(imw / down)
        batch_imgs = np.zeros((batch_size, 3, imh, imw), dtype=np.float32)
        batch_center_heatmap = np.zeros((batch_size, self.NUM_CLASSES, hm_h, hm_w), dtype=np.float32)
        batch_corner_heatmap = np.zeros((batch_size, self.NUM_CLASSES, hm_h, hm_w), dtype=np.float32)
        batch_offsets = np.zeros((batch_size, self.NUM_REG_PARAMS, hm_h, hm_w), dtype=np.float32)
        batch_offsets_mask = np.zeros((batch_size, 1, hm_h, hm_w), dtype=np.float32)
        for i, (img_path, objs) in enumerate(batch_annots):
            img = cv2.imread(img_path)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            img, objs = self.data_augmentation(img, objs)
            img = np.array(img)
            img = img / 255
            img = ut.normalize(img)
            batch_imgs[i, ...] = np.transpose(img, (2, 0, 1))
            for points, cls in objs:
                points = np.array(points)
                center, regs = self.transform_points(points)
                cx, cy = (center / down).astype(np.int)
                ut.draw_points_heatmap(batch_center_heatmap[i, cls], [[cx, cy]])
                ut.draw_points_heatmap(batch_corner_heatmap[i, cls], (points / 4).astype(np.int))
                batch_offsets[i, :, cy, cx] = regs
                batch_offsets_mask[i, 0, cy, cx] = 1
        batch_imgs = torch.from_numpy(batch_imgs).to(self.DEVICE)
        batch_center_heatmap = torch.from_numpy(batch_center_heatmap).to(self.DEVICE)
        batch_corner_heatmap = torch.from_numpy(batch_corner_heatmap).to(self.DEVICE)
        batch_offsets = torch.from_numpy(batch_offsets).to(self.DEVICE)
        batch_offsets_mask = torch.from_numpy(batch_offsets_mask).to(self.DEVICE)
        batch_labels = dict(
            center_heatmap=batch_center_heatmap,
            corner_heatmap=batch_corner_heatmap,
            offsets=batch_offsets,
            offsets_mask=batch_offsets_mask,
        )
        return batch_imgs, batch_labels


class CircleDetDataset(ShapeDetDataset):
    REG_MODE = 'circle'
    NUM_REG_PARAMS = 1
class CenterDetDataset(ShapeDetDataset):
    REG_MODE = 'centernet'
    NUM_REG_PARAMS = 2
class FCOSDetDataset(ShapeDetDataset):
    REG_MODE = 'fcos'
    NUM_REG_PARAMS = 4
class PolygonDetDataset(ShapeDetDataset):
    REG_MODE = 'offsets'






import itertools
import torch
import torch.nn as nn
import numpy as np


class BBoxTransform(nn.Module):
    def forward(self, anchors, regression):
        """
        decode_box_outputs adapted from https://github.com/google/automl/blob/master/efficientdet/anchors.py

        Args:
            anchors: [batchsize, boxes, (y1, x1, y2, x2)]
            regression: [batchsize, boxes, (dy, dx, dh, dw)]

        Returns:

        """
        y_centers_a = (anchors[..., 0] + anchors[..., 2]) / 2
        x_centers_a = (anchors[..., 1] + anchors[..., 3]) / 2
        ha = anchors[..., 2] - anchors[..., 0]
        wa = anchors[..., 3] - anchors[..., 1]

        w = regression[..., 3].exp() * wa
        h = regression[..., 2].exp() * ha

        y_centers = regression[..., 0] * ha + y_centers_a
        x_centers = regression[..., 1] * wa + x_centers_a

        ymin = y_centers - h / 2.
        xmin = x_centers - w / 2.
        ymax = y_centers + h / 2.
        xmax = x_centers + w / 2.

        return torch.stack([xmin, ymin, xmax, ymax], dim=2)


class ClipBoxes(nn.Module):

    def __init__(self):
        super(ClipBoxes, self).__init__()

    def forward(self, boxes, img):
        batch_size, num_channels, height, width = img.shape

        boxes[:, :, 0] = torch.clamp(boxes[:, :, 0], min=0)
        boxes[:, :, 1] = torch.clamp(boxes[:, :, 1], min=0)

        boxes[:, :, 2] = torch.clamp(boxes[:, :, 2], max=width - 1)
        boxes[:, :, 3] = torch.clamp(boxes[:, :, 3], max=height - 1)

        return boxes


class Anchors(nn.Module):
    """
    adapted and modified from https://github.com/google/automl/blob/master/efficientdet/anchors.py by Zylo117
    """

    def __init__(self, anchor_scale=4., pyramid_levels=None, **kwargs):
        super().__init__()
        self.anchor_scale = anchor_scale

        if pyramid_levels is None:
            self.pyramid_levels = [3, 4, 5, 6, 7]

        self.strides = kwargs.get('strides', [2 ** x for x in self.pyramid_levels])
        self.scales = np.array(kwargs.get('scales', [2 ** 0, 2 ** (1.0 / 3.0), 2 ** (2.0 / 3.0)]))
        self.ratios = kwargs.get('ratios', [(1.0, 1.0), (1.4, 0.7), (0.7, 1.4)])

        self.last_anchors = {}
        self.last_shape = None

    def forward(self, image, dtype=torch.float32):
        """Generates multiscale anchor boxes.

        Args:
          image_size: integer number of input image size. The input image has the
            same dimension for width and height. The image_size should be divided by
            the largest feature stride 2^max_level.
          anchor_scale: float number representing the scale of size of the base
            anchor to the feature stride 2^level.
          anchor_configs: a dictionary with keys as the levels of anchors and
            values as a list of anchor configuration.

        Returns:
          anchor_boxes: a numpy array with shape [N, 4], which stacks anchors on all
            feature levels.
        Raises:
          ValueError: input size must be the multiple of largest feature stride.
        """
        image_shape = image.shape[2:]

        if image_shape == self.last_shape and image.device in self.last_anchors:
            return self.last_anchors[image.device]

        if self.last_shape is None or self.last_shape != image_shape:
            self.last_shape = image_shape

        if dtype == torch.float16:
            dtype = np.float16
        else:
            dtype = np.float32

        boxes_all = []
        for stride in self.strides:
            boxes_level = []
            for scale, ratio in itertools.product(self.scales, self.ratios):
                if image_shape[1] % stride != 0:
                    raise ValueError('input size must be divided by the stride.')
                base_anchor_size = self.anchor_scale * stride * scale
                anchor_size_x_2 = base_anchor_size * ratio[0] / 2.0
                anchor_size_y_2 = base_anchor_size * ratio[1] / 2.0

                x = np.arange(stride / 2, image_shape[1], stride)
                y = np.arange(stride / 2, image_shape[0], stride)
                xv, yv = np.meshgrid(x, y)
                xv = xv.reshape(-1)
                yv = yv.reshape(-1)

                # y1,x1,y2,x2
                boxes = np.vstack((yv - anchor_size_y_2, xv - anchor_size_x_2,
                                   yv + anchor_size_y_2, xv + anchor_size_x_2))
                boxes = np.swapaxes(boxes, 0, 1)
                boxes_level.append(np.expand_dims(boxes, axis=1))
            # concat anchors on the same level to the reshape NxAx4
            boxes_level = np.concatenate(boxes_level, axis=1)
            boxes_all.append(boxes_level.reshape([-1, 4]))

        anchor_boxes = np.vstack(boxes_all)

        anchor_boxes = torch.from_numpy(anchor_boxes.astype(dtype)).to(image.device)
        anchor_boxes = anchor_boxes.unsqueeze(0)

        # save it for later use to reduce overhead
        self.last_anchors[image.device] = anchor_boxes
        return anchor_boxes
